import os
import sys
import textwrap
import time

import humanize
import oyaml as yaml

from cloudmesh.burn.image import Image
from cloudmesh.burn.usb import USB
from cloudmesh.burn.util import os_is_linux
from cloudmesh.burn.util import os_is_mac
from cloudmesh.burn.util import os_is_pi
from cloudmesh.burn.util import os_is_windows
from cloudmesh.common.Shell import Shell
from cloudmesh.common.Shell import windows_not_supported
from cloudmesh.common.Tabulate import Printer
from cloudmesh.common.console import Console
from cloudmesh.common.sudo import Sudo
from cloudmesh.common.systeminfo import get_platform
from cloudmesh.common.util import banner
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import readfile as common_readfile
from cloudmesh.common.util import yn_choice


# noinspection PyBroadException
def _execute(msg, command):
    Console.ok(msg)
    try:
        os.system(command.strip())
    except:  # noqa: E722
        Console.error("{command} failed")
        # but ignore error


def location(host_os=None, card_os="raspberry", volume="boot"):
    """
    Returns the location of the specific volume after mounting

    @param volume: either boot or root
    @type volume: str
    @param host_os: The OS on which this command is run and the mount occurs.
                    Values are mac, raspberry, ubuntu, linux. Linux maps to ubuntu
    @type host_os: str
    @param card_os: The operating system to be burned on the card.
                    Values are raspberry, ubuntu
    @type card_os: str
    @return:
    @rtype:
    """

    # for backwards compatibility map linux to ubuntu
    if host_os.lower() in ["linux"]:
        host_os = "ubuntu"

    user = os.environ.get('USER')

    # where [host_os][burn_os][volume]
    where = yaml.safe_load(textwrap.dedent(f"""
            raspberry:
              raspberry:
                root: /media/{user}/rootfs
                boot: /media/{user}/boot
              ubuntu:
                root: /media/{user}/writable
                boot: /media/{user}/system-boot
            macos:
              raspberry:
                root: /Volumes/rootfs
                boot: /Volumes/boot
              ubuntu:
                root: /Volumes/writable
                boot: /Volumes/system-boot
            ubuntu:
              raspberry:
                root: /media/{user}/rootfs
                boot: /media/{user}/boot
              ubuntu:
                root: /media/{user}/writable
                boot: /media/{user}/system-boot
            linux:
              raspberry:
                root: /media/{user}/rootfs
                boot: /media/{user}/boot
              ubuntu:
                root: /media/{user}/writable
                boot: /media/{user}/system-boot
            """))
    try:
        return where[host_os][card_os][volume]
    except Exception as e:
        print(e)
        return "os_undefined_in_location"


class SDCard:

    def __init__(self, card_os=None, host_os=None):
        """
        Creates mount point strings based on OS and the host where it is executed

        :param card_os: the os that is part of the mount. Default: raspberry
        :type card_os: str
        :param host_os: the host on which we execute the command
        :type host_os: possible values: raspberry, macos, linux
        """
        self.card_os = card_os or "raspberry"
        self.host_os = host_os or get_platform()

    @staticmethod
    def execute(command=None, decode="True", debug=False):
        """
        Executes the command

        :param command: The command to run
        :type command: list or str
        :param decode:
        :type decode:
        :param debug:
        :type debug:
        :return:
        :rtype:
        """
        result = Sudo.execute(command, decode=decode, debug=debug)
        return result

    @property
    def root_volume(self):
        """
        the location of system volume on the SD card for the specified host
        and os in Location initialization

        TODO: not implemented

        :return: the location
        :rtype: str
        """
        return location(volume="root", card_os=self.card_os, host_os=self.host_os)

    @property
    def boot_volume(self):
        """
        the location of the boot volume for the specified host and os in
        Location initialization

        :return: the location
        :rtype: str
        """
        return location(volume="boot", card_os=self.card_os, host_os=self.host_os)

    def ls(self):
        """
        List all file systems on the SDCard. This is for the PI rootfs and boot

        :return: A dict representing the file systems on the SDCCards
        :rtype: dict
        """

        r = Shell.run("mount -l").splitlines()
        root_fs = self.root_volume
        boot_fs = self.boot_volume

        details = {}
        for line in r:
            if str(root_fs) in line or str(boot_fs) in line:
                entry = \
                    line.replace(" on ", "|") \
                        .replace(" type ", "|") \
                        .replace(" (", "|") \
                        .replace(") [", "|") \
                        .replace("]", "") \
                        .split("|")
                # print(entry)
                detail = {
                    "device": entry[0],
                    "path": entry[1],
                    "type": entry[2],
                    "parameters": entry[3],
                    "name": entry[4],
                }
                details[detail["name"]] = detail
        return details

    @staticmethod
    def _readfile(filename=None, split=False, trim=False, decode=True):
        """
        Reads the content of the file as sudo and returns the result

        :param filename: the filename
        :type filename: str
        :param split: if true returns a list of lines
        :type split: bool
        :param trim: trim trailing whitespace. This is useful to
                     prevent empty string entries when splitting by '\n'
        :type trim: bool
        :param decode:
        :type decode: bool
        :return: the content
        :rtype: str or list
        """
        Sudo.execute("sync")

        if os_is_mac():
            if decode:
                mode = "r"
            else:
                mode = "rb"
            content = common_readfile(filename, mode=mode)
        else:
            Sudo.password()
            result = Sudo.execute(f"cat {filename}", decode=decode)
            content = result.stdout

        if trim:
            content = content.rstrip()

        if split:
            content = content.splitlines()
        Sudo.execute("sync")

        return content

    @staticmethod
    def readfile(filename=None, split=False, trim=False, decode=True):
        """
        Reads the content of the file as sudo and returns the result

        :param filename: the filename
        :type filename: str
        :param split: if true returns a list of lines
        :type split: bool
        :param trim: trim trailing whitespace. This is useful to
                     prevent empty string entries when splitting by '\n'
        :type trim: bool
        :param decode:
        :type decode: bool
        :return: the content
        :rtype: str or list
        """

        found = False
        for i in range(0, 10):
            try:
                time.sleep(0.5)
                content = SDCard._readfile(filename=filename, trim=trim, decode=decode, split=split)
                found = content != []
                if found:
                    break
            except:  # noqa
                pass
            print(f"read {filename} attempt {i}")

        return content

    @staticmethod
    def writefile(filename=None, content=None, append=False):
        """
        Writes the content in the the given file.

        :param filename: the filename
        :type filename: str
        :param content: the content
        :type content: str
        :param append: if true it append it at the end, otherwise the file will
                       be overwritten
        :type append: bool
        :return: the output created by the write process
        :rtype: int
        """
        os.system("sync")
        if append:
            content = Sudo.readfile(filename, split=False, decode=True) + content
        command = f"echo '{content}' | sudo cp /dev/stdin {filename}"
        if os_is_mac() and "\0" in command:
            command = command.replace("\0", "")
        os.system(command)
        os.system("sync")

        return content

    # noinspection PyBroadException
    @staticmethod
    def size(device="/dev/sdX"):

        size = 64 * 1000 ** 3  # this is a bug as we need that for Linux and PI

        if os_is_mac():
            result = Shell.run("diskutil list external").splitlines()
            for line in result:
                if "FDisk_partition_scheme" in line:
                    data = line.split()
                    size, unit = data[2].replace("*", ""), data[3]
                    if unit == "GB":
                        size = int(float(size) * 1000 ** 3)
                    else:
                        size = 64
                        Console.error("Unit not GB")
                    break
        elif os_is_linux():
            try:
                result = Shell.run(f"sudo blockdev --getsize64 {device}").strip()
                result = int(result)
            except Exception as e:  # noqa: F841
                Console.error(f"Could not determine size of the device {device}")
                sys.exit()
        return size

    @windows_not_supported
    def format_device(self,
                      device='dev/sdX',
                      unmount=True,
                      yes=False,
                      verbose=True):
        """
        Formats device with one FAT32 partition

        WARNING: make sure you have the right device, this command could
                 potentially erase your OS

        :param device: The device on which we format
        :type device: str
        :param unmount:
        :type unmount:
        :param yes:
        :type yes:
        :param verbose:
        :type verbose:
        :return:
        :rtype:
        """

        _title = "UNTITLED"

        def prepare_sdcard():
            """
            ensures a card is detected and unmounted
            :return: True if prepared
            :rtype: bool
            """
            #
            Console.ok(f'sudo eject -t {device}')
            os.system(f'sudo eject -t {device}')
            time.sleep(3)
            device_basename = os.path.basename(device)
            result = Shell.run('lsblk')
            if device_basename in result.split():
                for line in result.splitlines():
                    line = line.split()
                    if device_basename in line[0] and len(line) > 6:
                        Console.ok(f'sudo umount {line[6]}')
                        os.system(f'sudo umount {line[6]}')
                return True
            else:
                Console.error("SD Card not detected. Please reinsert "
                              "card reader. ")
                if not yn_choice("Card reader re-inserted? No to cancel "
                                 "operation"):
                    return False
                else:
                    time.sleep(3)
                    return prepare_sdcard()

        Sudo.password()
        if os_is_linux() or os_is_pi():

            if verbose:
                banner(f"format {device}")
            else:
                print(f"format {device}")

            if not prepare_sdcard():
                return False

            # TODO Gregor verify commenting out the below is ok
            # if os_is_mac():
            #    self.mount(device=device)

            user = os.environ.get('USER')

            script = textwrap.dedent(f"""
                ls /media/{user}
                sudo parted {device} --script -- mklabel msdos
                sudo parted {device} --script -- mkpart primary fat32 1MiB 100%
                sudo mkfs.vfat -n {_title} -F32 {device}1
                sudo parted {device} --script print""").strip().splitlines()
            for line in script:
                _execute(line, line)

            os.system("sudo sync")
            if unmount:
                self.unmount(device=device)  # without dev we unmount but do not eject. If
                # we completely eject, burn will fail to detect the device.
                os.system("sudo sync")

            Console.ok("Formatted SD Card")

        elif os_is_mac():

            details = USB.get_dev_from_diskutil()

            # checking if string contains list element
            valid = any(entry in device for entry in details)

            if not valid:
                Console.error(f"this device can not be used for formatting: {device}")
                return False

            elif len(details) > 1:
                Console.error("For security reasons, please only put one USB writer in")
                Console.msg(f"we found {details}")
                return False

            else:

                details = USB.get_from_diskutil()

                USB.print_details(details)

                print()
                if yes or yn_choice(f"\nDo you like to format {device} as {_title}"):
                    _execute(f"Formatting {device} as {_title}",
                             f"sudo diskutil eraseDisk FAT32 {_title} MBRFormat {device}")

        else:
            raise NotImplementedError("Not implemented for this OS")

        return True

    def _info(self):
        print("root", self.root_volume)
        print("boot", self.boot_volume)

    def probe_os(self):
        if os_is_mac():
            where_raspberry = location(host_os="macos", card_os="raspberry", volume="boot")
            where_ubuntu = location(host_os="macos", card_os="ubuntu", volume="boot")
        elif os_is_pi():
            where_raspberry = location(host_os="raspberry", card_os="raspberry", volume="boot")
            where_ubuntu = location(host_os="raspberry", card_os="ubuntu", volume="boot")
        elif os_is_linux():
            where_raspberry = location(host_os="ubuntu", card_os="raspberry", volume="boot")
            where_ubuntu = location(host_os="ubuntu", card_os="ubuntu", volume="boot")
        else:
            return "unkwn"

        if os.path.exists(where_raspberry):
            return "raspberryos"
        elif os.path.exists(where_ubuntu):
            return "ubuntu"
        else:
            return "unkown"

    # @windows_not_supported
    # TODO Gregor verify the default arg for card_os is ok
    def mount(self, device=None, card_os="raspberry"):
        """
        Mounts the current SD card
        """

        if os_is_linux():
            Sudo.password()
            dmesg = USB.get_from_dmesg()

            # TODO Need a better way to identify which sd card to use for mounting
            # instead of iterating over all of them

            for usbcard in dmesg:

                dev = device or usbcard['dev']
                print(f"Mounting filesystems on {dev}")
                try:
                    Console.ok(f"mounting {device}")
                    os.system('sudo sync')  # flush any pending/in-process writes
                    os.system(f"sudo eject -t {device}")
                    os.system('sudo sync')  # flush any pending/in-process writes

                    # ensure the card is mounted before returning
                    device_basename = os.path.basename(device)
                    part1 = False
                    part2 = False
                    for i in range(10):
                        result = Shell.run('lsblk')
                        if device_basename in result.split():
                            for line in result.splitlines():
                                line = line.split()
                                if device_basename + '1' in line[0] and len(line) > 6:
                                    part1 = True
                                elif device_basename + '2' in line[0] and len(line) > 6:
                                    part2 = True
                        if part1 and part2:
                            # card is fully mounted
                            break
                        time.sleep(0.5)

                    if not part1 and not part2:
                        raise Exception("card failed to mount both partitions")

                except Exception as e:
                    print(e)

        elif os_is_pi():
            Sudo.password()

            if card_os is None:
                Console.error("Please specify the OS you have on the SD Card")
                return ""
            self.card_os = card_os
            dmesg = USB.get_from_dmesg()
            print(dmesg)

            # TODO Need a better way to identify which sd card to use for mounting
            # instead of iterating over all of them

            os.system('sudo sync')  # flush any pending/in-process writes

            for usbcard in dmesg:

                dev = device or usbcard['dev']
                print(f"Mounting filesystems on {dev} assuming it is {card_os} as you specified")
                sd1 = f"{dev}1"
                sd2 = f"{dev}2"
                try:
                    if os.path.exists(sd1):
                        Console.ok(f"mounting {sd1} {self.boot_volume}")
                        os.system(f"sudo mkdir -p {self.boot_volume}")
                        os.system(f"sudo mount -t vfat {sd1} {self.boot_volume}")
                except Exception as e:
                    print(e)
                try:
                    if os.path.exists(sd2):
                        Console.ok(f"mounting {sd2} {self.root_volume}")
                        os.system(f"sudo mkdir -p {self.root_volume}")
                        os.system(f"sudo mount -t ext4 {sd2} {self.root_volume}")
                except Exception as e:
                    print(e)

        elif os_is_mac():

            command = f"diskutil mountDisk {device}"
            print(command)
            os.system(command)

        else:
            Console.error("Not yet implemented for your OS")

        Sudo.execute("sync")
        return ""

    @windows_not_supported
    def unmount(self, device=None, card_os="raspberry", full=False):
        """
        Unmounts the current SD card. param full indicates whether to use -t flag

        :param device: device to unmount, e.g. /dev/sda
        :type device: str
        :param card_os:
        :type card_os:
        :param full:
        :type full:
        :return:
        :rtype:
        """

        # self.card_os = card_os

        os.system('sudo sync')  # flush any pending/in-process writes

        os.system("sync")
        if os_is_linux() or os_is_pi():
            Sudo.password()
            if full:
                _execute(f"eject {device}", f"sudo eject {device}")
            else:
                # _execute(f"eject {device}", f"sudo eject -t {device}")
                device_basename = os.path.basename(device)
                result = Shell.run('lsblk')
                if device_basename in result.split():
                    for line in result.splitlines():
                        line = line.split()
                        if device_basename in line[0] and len(line) > 6:
                            Console.ok(f'sudo umount {line[6]}')
                            os.system(f'sudo umount {line[6]}')
            # _execute(f"unmounting {self.boot_volume}", f"sudo umount {self.boot_volume}")
            # _execute(f"unmounting  {self.root_volume}", f"sudo umount {self.root_volume}")
        elif os_is_mac():

            _execute(f"unmounting {self.boot_volume}", f"diskutil umountDisk {device}")

        else:
            Console.error("Not yet implemented for your OS")
            return ""
        os.system("sync")
        # rm = [f"sudo rmdir {self.boot_volume}",
        #      f"sudo rmdir {self.root_volume}"]

        # for command in rm:
        #    _execute(command, command)

        return True

    @windows_not_supported
    def load_device(self, device='dev/sdX'):
        """
        Loads the USB device via trayload

        :param device: The device on which we format
        :type device: str
        """
        if os_is_linux() or os_is_pi():
            banner(f"load {device}")
            Sudo.password()
            os.system(f"sudo eject -t {device}")

        else:
            raise Console.error("Not implemented for this OS")

    @windows_not_supported
    def backup(self, device=None, to_file=None, blocksize="4m"):
        if device is None:
            Console.error("Device must have a value")
        if to_file is None:
            Console.error("To file must have a value")
        else:
            Sudo.password()

            to_file = path_expand(to_file)

            size = SDCard.size(device)

            to_file = path_expand(to_file)

            #
            # speed up burning on MacOS
            #
            if device.startswith("/dev/disk"):
                device = device.replace("/dev/disk", "/dev/rdisk")

            command = f"sudo dd if={device} bs={blocksize} |" \
                      f' tqdm --bytes --total {size} --ncols 80|' \
                      f"dd of={to_file} bs={blocksize}"

            print()
            Console.info(command)
            print()

            os.system(command)

    @windows_not_supported
    def burn_sdcard(self,
                    image=None,
                    tag=None,
                    device=None,
                    blocksize="4M",
                    name="the inserted card",
                    yes=False):
        """
        Burns the SD Card with an image

        :param image: Image object to use for burning (used by copy)
        :type image: str
        :param name:
        :type name: str
        :param tag: tag object used for burning (used by sdcard)
        :type tag: str
        :param device: Device to burn to, e.g. /dev/sda
        :type device: str
        :param blocksize: the blocksize used when writing, default 4M
        :type blocksize: str
        :param yes:
        :type yes: str

        """
        if image and tag:
            Console.error("Implementation error, burn_sdcard can't have image "
                          "and tag.")
            return ""

        Console.info(f"Burning {name} ...")
        if image is not None:
            image_path = image
        else:
            image = Image().find(tag=tag)

            if image is None:
                Console.error("No matching image found.")
                return ""
            elif len(image) > 1:
                Console.error("Too many images found")
                print(Printer.write(image,
                                    order=["tag", "version"],
                                    header=["Tag", "Version"]))
                return ""

            image = image[0]

            if "ubuntu" in image["url"]:
                _name = os.path.basename(Image.get_name(image["url"]))
                _name = _name.replace(".xz", "")
            else:
                _name = os.path.basename(Image.get_name(image["url"])) + ".img"

            image_path = Image().directory + "/" + _name

            print(image_path)

            if not os.path.isfile(image_path):
                tags = ' '.join(tag)

                print()
                Console.error(f"Image with tags '{tags}' not found. To download use")
                print()
                Console.msg(f"cms burn image get {tags}")
                print()
                return ""

        orig_size = size = humanize.naturalsize(os.path.getsize(image_path))

        # size = details[0]['size']
        n, unit = size.split(" ")
        unit = unit.replace("GB", "G")
        unit = unit.replace("MB", "M")
        n = float(n)
        if unit == "G":
            n = n * 1000 ** 3
        elif unit == "M":
            n = n * 1000 ** 2
        size = int(n)

        banner(f"Preparing the SDCard {name}")
        print(f"Name:       {name}")
        print(f"Image:      {image_path}")
        print(f"Image Size: {orig_size}")
        print(f"Device:     {device}")
        print(f"Blocksize:  {blocksize}")

        if os_is_mac():
            blocksize = blocksize.lower()

        print()

        Sudo.password()

        if device is None:
            Console.error("Please specify a device")
            return

        #
        # speedup burn for MacOS
        #
        if device.startswith("/dev/disk"):
            device = device.replace("/dev/disk", "/dev/rdisk")

        if os_is_mac():
            details = USB.get_from_diskutil()
            USB.print_details(details)

        if not (yes or yn_choice(f"\nDo you like to write {name} on {device} "
                                 f"with the image {image_path}")):
            return ""

        # TODO Gregor verify this is ok commenting out this line
        # self.mount(device=device)

        if os_is_mac():
            command = f"sudo dd if={image_path} bs={blocksize} |" \
                      f' tqdm --bytes --total {size} --ncols 80 |' \
                      f" sudo dd of={device} bs={blocksize}"
        else:
            # command = f"sudo dd if={image_path} of={device} bs={blocksize} status=progress conv=fsync"
            command = f"sudo dd if={image_path} bs={blocksize} oflag=direct |" \
                      f' tqdm --bytes --total {size} --ncols 80 |' \
                      f" sudo dd of={device} bs={blocksize} iflag=fullblock " \
                      f"oflag=direct conv=fsync"
        print(command)
        os.system(command)

        Sudo.execute("sync")
        if os_is_linux():
            self.unmount(device=device, full=True)
        else:
            self.unmount(device=device)

    def copy(self, device=None, from_file="latest"):
        if device is None:
            Console.error("Device must have a value")
        self.burn_sdcard(image=from_file, device=device)

    @windows_not_supported
    def info(self,
             print_os=True,
             print_fdisk=True,
             print_stdout=True,
             output="table"):
        """
        Finds out information about USB devices

        TODO: should we rename print_stdout to debug? seems more in
              line with cloudmesh

        :param print_os:
        :type print_os:
        :param print_fdisk:
        :type print_fdisk:
        :param print_stdout: if set to True prints debug information
        :type print_stdout: bool
        :param output:
        :type output:
        :return: dict with details about the devices
        :rtype: dict
        """

        if print_os and print_stdout:
            if os_is_pi():
                banner("This is  Raspberry PI")
            elif os_is_mac():
                banner("This is Mac")
            elif os_is_windows():
                banner("This is a Windows Computer")
            elif os_is_linux():
                banner("This is a Linux Computer")
            else:
                Console.error("unkown OS")
                sys.exit(1)

        if os_is_pi() and print_fdisk and print_stdout:
            result = USB.fdisk("/dev/mmcblk0")
            if print_stdout:
                banner("Operating System SD Card")
                print(result)

        details = USB.get_from_usb()

        if print_stdout and details is not None:
            banner("USB Device Probe")
            print(Printer.write(
                details,
                order=["address",
                       "bus",
                       "idVendor",
                       "idProduct",
                       "hVendor",
                       "hProduct",
                       "iManufacturer",
                       "iSerialNumber",
                       "usbVersion",
                       "comment"],
                header=["Adr.",
                        "bus",
                        "Vendor",
                        "Prod.",
                        "H Vendor",
                        "H Prod.",
                        "Man.",
                        "Ser.Num.",
                        "USB Ver.",
                        "Comment"],
                output=output)
            )

        # devices = USB.get_devices()

        # banner("Devices found")

        # print ('\n'.join(sorted(devices)))

        if os_is_mac():

            names = USB.get_dev_from_diskutil()

            details = USB.get_from_diskutil()
        else:
            details = USB.get_from_dmesg()

        if print_stdout:
            banner("SD Cards Found")

            if os_is_mac():
                print("We found the follwing cards:")
                print("  - /dev/" + "\n  - /dev/".join(names))
                print()
                print("We found the follong file systems on these disks:")
                print()

            print(Printer.write(details,
                                order=[
                                    "dev",
                                    "info",
                                    "formatted",
                                    "size",
                                    "active",
                                    "readable",
                                    "empty",
                                    "direct-access",
                                    "removable",
                                    "writeable"],
                                header=[
                                    "Path",
                                    "Info",
                                    "Formatted",
                                    "Size",
                                    "Plugged-in",
                                    "Readable",
                                    "Empty",
                                    "Access",
                                    "Removable",
                                    "Writeable"],
                                output=output)
                  )

            # lsusb = USB.get_from_lsusb()
            # from pprint import pprint
            # pprint (lsusb)

            # endors = USB.get_vendor()
            # print(vendors)

            # udev = subprocess.getoutput("udevadm info -a -p  $(udevadm info -q path -n /dev/sda)")
            #
            # attributes = ["vendor","model", "model", "version", "manufacturer",
            #     "idProduct", "idVendor"]
            # for line in udev.splitlines():
            #    if any(word in line for word in attributes):
            #        print(line)

        if print_stdout:

            if os_is_linux():
                card = SDCard()
                m = card.ls()

                banner("Mount points")
                if len(m) != 0:
                    print(Printer.write(m,
                                        order=["name", "path", "type", "device", "parameters"],
                                        header=["Name", "Path", "Type", "Device", "Parameters"],
                                        output=output))
                else:
                    Console.warning("No mount points found. Use cms burn mount")
                    print()

        # Convert details into a dict where the key for each entry is the device
        details = {detail['dev']: detail for detail in details}

        return details



import textwrap

from cloudmesh.common.Shell import Shell
from cloudmesh.common.util import readfile


class Cmdline:

    def __init__(self):
        # self.script = " ".join(textwrap.dedent("""
        # console=serial0,115200
        # console=tty1
        # root=PARTUUID=904a3764-02
        # rootfstype=ext4
        # elevator=deadline
        # fsck.repair=yes
        # rootwait
        # quiet
        # init=/usr/lib/raspi-config/init_resize.sh
        # splash
        # plymouth.ignore-serial-consoles
        # systemd.run=/boot/firstrun.sh
        # systemd.run_success_action=reboot
        # systemd.unit=kernel-command-line.target
        # """).splitlines()).strip()

        # Commented out above since we should just append
        # the lines below to the existing cmdline.txt since
        # root PARTUUID may vary

        # self.cmdline will be populated when .read() is called
        self.cmdline = None
        # the space-separated values to add to the end of cmdline
        self.script = " ".join(textwrap.dedent("""
        splash
        plymouth.ignore-serial-consoles
        systemd.run=/boot/firstrun.sh
        systemd.run_success_action=reboot
        systemd.unit=kernel-command-line.target
        """).splitlines()).strip()

    def read(self, filename=None):
        """
        Read a pre-existing cmdline.txt and store it
        """
        if filename is None:
            raise Exception("read called with no filename")
        self.cmdline = readfile(filename).strip()

    def write(self, filename=None):
        """
        Write the cmdline config to the specified filename
        """
        if self.cmdline is None:
            # Cmdline varies by burn
            raise Exception("Please read a pre-existing cmdline.txt first")
        if filename is None:
            raise Exception("write called with no filename")
        Shell.run(f'echo "{self.cmdline} {self.script}" | sudo tee {filename}')

    def get(self):
        """
        Return the proper cmdline with the necessary commands
        """
        if self.cmdline is None:
            print("Using example cmdline.txt. Not safe for usage. For testing only")
            self.cmdline = self._example()

        return self.cmdline + self.script

    def _example(self):
        """
        An example cmdline.txt for testing purposes
        """
        return " ".join(textwrap.dedent("""
        console=serial0,115200
        console=tty1
        root=PARTUUID=904a3764-02
        rootfstype=ext4
        elevator=deadline
        fsck.repair=yes
        rootwait
        quiet
        init=/usr/lib/raspi-config/init_resize.sh
        """).splitlines()).strip()



"""
Implementation of a function the set the WIFI configuration.
This function is primarily developed for a Raspberry PI
"""
import textwrap

from cloudmesh.common.console import Console
from cloudmesh.common.sudo import Sudo
from cloudmesh.common.util import writefile


class Wifi:
    """
    The class is used to group a number of useful variables and functions so
    it is easier to program and manage Wifi configurations.

    The default location for the configuration file is

    /etc/wpa_supplicant/wpa_supplicant.conf

    """

    location = "/etc/wpa_supplicant/wpa_supplicant.conf"

    template = textwrap.dedent("""
        wifis:
          wlan0:
          dhcp4: true
          optional: true
          access-points:
            {ssid}:
              password: "{password}"
    """).strip()

    @staticmethod
    def set(ssid=None,
            password=None,
            country="US",
            psk=True,
            location=location,
            sudo=False):
        """
        Sets the wifi. Only works for psk based wifi

        :param ssid: The ssid
        :type ssid: str
        :param password: The password
        :type password: str
        :param country: Two digit country code
        :type country: str
        :param psk: If true uses psk authentication
        :type psk: bool
        :param location: The file where the configuration file should be written to
        :type location: str
        :param sudo: If tru the write will be done with sudo
        :type sudo: bool
        :return: True if success
        :rtype: bool
        """

        if ssid is None or password is None:
            Console.error("SSID or password not set")
            return False

        config = Wifi.template.format(**locals())

        try:
            if sudo:
                Sudo.writefile(location, config)
            else:
                writefile(location, config)

        except FileNotFoundError as e:  # noqa: F841
            Console.error(f"The file does not exist: {location}")
            return False
        return True



from abc import ABC, abstractmethod

from cloudmesh.common.parameter import Parameter


class AbstractBurner(ABC):

    # @abstractmethod
    def cluster(self, arguments=None):
        """
        burns a list of cards from the inventory

        :param arguments:
        :type arguments:
        :return:
        :rtype:
        """
        burning = Parameter.expand(arguments.burning)
        for host in burning:
            arguments.burning = burning
            self.burn(arguments)

    @abstractmethod
    def burn(self, arguments=None):
        """
        burns a single card from the inventory

        :param arguments:
        :type arguments:
        :return:
        :rtype:
        """
        raise NotImplementedError

    def inventory(self, arguments=None):
        """
        Creates the inventory for a cluster

        :param arguments:
        :type arguments:
        :return:
        :rtype:
        """
        raise NotImplementedError



import os
import textwrap
import zipfile
from pathlib import Path

import oyaml as yaml
import requests
import urllib3
from cloudmesh.burn.util import sha1sum
from cloudmesh.burn.util import sha256sum
from cloudmesh.common.Tabulate import Printer
from cloudmesh.common.console import Console
from cloudmesh.common.util import banner
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import readfile, writefile

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class Ubuntu:
    download = "https://ubuntu.com/download/raspberry-pi/thank-you"
    release = "https://cdimage.ubuntu.com/releases/"
    distribution = [
        {
            "version": "20.04.2&architecture=server-arm64+raspi",
            "date": "2021-02-01",
            "type": "ubuntu",
            "os": "ubuntu",
            "tag": "ubuntu-20.04.2-64-bit",
            "description": "Ubuntu Server 20.04.2 LTS 64-bit",
            "alt": f"{download}?version=20.04.2&architecture=server-arm64+raspi",
            "url": f"{release}/20.04.2/release/ubuntu-20.04.2-preinstalled-server-arm64+raspi.img.xz",
            "sha256": "31884b07837099a5e819527af66848a9f4f92c1333e3ef0693d6d77af66d6832"
            # check
            # echo "31884b07837099a5e819527af66848a9f4f92c1333e3ef0693d6d77af66d6832
            #      *ubuntu-20.04.2-preinstalled-server-arm64+raspi.img.xz" | shasum -a 256 --check
        },
        {
            "version": "20.04.2&architecture=server-armhf+raspi",
            "date": "2021-02-01",
            "type": "ubuntu",
            "os": "ubuntu",
            "tag": "ubuntu-20.04.2-32-bit",
            "description": "Ubuntu Server 20.04.2 LTS 32-bit",
            "alt": f"{download}?version=20.04.2&architecture=server-armhf+raspi",
            "url": f"{release}/20.04.2/release/ubuntu-20.04.2-preinstalled-server-armhf+raspi.img.xz",
            "sha256": "7b348d080648b8e36e1f29671afd973a0878503091b935b69828f2c7722dfb58"
            # check
            # echo "7b348d080648b8e36e1f29671afd973a0878503091b935b69828f2c7722dfb58
            #      *ubuntu-20.04.2-preinstalled-server-armhf+raspi.img.xz" | shasum -a 256 --check
        },

        {
            "version": "20.10&architecture=server-arm64+raspi",
            "date": "2021-02-01",
            "type": "ubuntu",
            "os": "ubuntu",
            "tag": "ubuntu-20.10-64-bit",
            "description": "Ubuntu Server 20.10 64-bit",
            "alt": f"{download}?version=20.10&architecture=server-arm64+raspi",
            "url": f"{release}/20.10/release/ubuntu-20.10-preinstalled-server-arm64+raspi.img.xz",
            "sha256": "73a98e83fbb6398f86902c7a83c826536f1afbd0be2d80cb0d7faa94ede33137"
            # check
            # echo "73a98e83fbb6398f86902c7a83c826536f1afbd0be2d80cb0d7faa94ede33137
            #      *ubuntu-20.10-preinstalled-server-arm64+raspi.img.xz" | shasum -a 256 --check
        },

        {
            "version": "20.10&architecture=server-armhf+raspi",
            "date": "2021-02-01",
            "type": "ubuntu",
            "os": "ubuntu",
            "tag": "ubuntu-20.10-32-bit",
            "description": "Ubuntu Server 20.10",
            "alt": f"{download}?version=20.10&architecture=server-armhf+raspi",
            "url": f"{release}/20.10/release/ubuntu-20.10-preinstalled-server-armhf+raspi.img.xz",
            "sha256": "941752f93313ba765069b48f38af18bbf961bebefcc9c4296a953cf1260f0fe1"
            # check
            # echo "941752f93313ba765069b48f38af18bbf961bebefcc9c4296a953cf1260f0fe1
            #      *ubuntu-20.10-preinstalled-server-armhf+raspi.img.xz" | shasum -a 256 --check
        },
        {
            "version": "20.10&architecture=desktop-arm64+raspi",
            "date": "2021-02-01",
            "type": "ubuntu",
            "os": "ubuntu",
            "tag": "ubuntu-desktop",
            "description": "Ubuntu Desktop 20.10 64-bit",
            "alt": f"{download}?version=20.10&architecture=desktop-arm64+raspi",
            "url": f"{release}/20.10/release/ubuntu-20.10-preinstalled-desktop-arm64+raspi.img.xz",
            "sha256": "2fa19fb53fe0144549ff722d9cd755d9c12fb508bb890926bfe7940c0b3555e8"
            # check
            # echo "2fa19fb53fe0144549ff722d9cd755d9c12fb508bb890926bfe7940c0b3555e8
            #      *ubuntu-20.10-preinstalled-desktop-arm64+raspi.img.xz" | shasum -a 256 --check
        },
    ]


# noinspection PyPep8
class Image(object):

    # self.directory: the folder where downloaded images are kept
    # self.image_name: the name of the image (or URL to fetch it from)
    # self.fullpath: the full path of the image, e.g.
    # /home/user/.cloudmesh/images/raspbian-2019.img

    def __init__(self):

        self.directory = os.path.expanduser('~/.cloudmesh/cmburn/images')
        self.cache = Path(os.path.expanduser("~/.cloudmesh/cmburn/distributions.yaml"))
        os.system('mkdir -p ' + self.directory)

        self.raspberry_images = {
            "lite": "https://downloads.raspberrypi.org/raspios_lite_armhf/images",
            "full": "https://downloads.raspberrypi.org/raspios_full_armhf/images"
        }
        # if name == 'latest':
        #    self.fullpath = self.directory + '/' + self.latest_version() + '.img'
        # else:
        #    self.fullpath = self.directory + '/' + self.image_name + '.img'

    def read_version_cache(self):
        """
        reads the image list cache from the default cache location

        :return:
        :rtype:
        """
        data = yaml.load(readfile(self.cache), Loader=yaml.SafeLoader)
        return data

    @staticmethod
    def find(tag=None):
        """
        finds image names in the image list with the given tags

        :param tag: list of strings
        :type tag: list
        :return: tag matiching images
        :rtype: dict
        """
        tag = tag or ['latest-lite']
        found = []
        data = Image.create_version_cache(refresh=False)
        for entry in data:
            match = True
            for t in tag:
                match = match and t in entry["tag"]
            if match:
                found.append(entry)
        if len(found) == 0:
            return None
        else:
            return found

    @staticmethod
    def create_version_cache(refresh=False):
        """
        creates a cache of all released pi images

        :param refresh: reresd it from the Web if True
        :type refresh: bool
        :return: writes it into ~/.cloudmesh/cmburn/distributions.yaml
        :rtype: file
        """

        data = {
            "lite": [],
            "full": []
        }
        cache = Path(os.path.expanduser("~/.cloudmesh/cmburn/distributions.yaml"))

        def fetch_kind(kind=None):
            print(f"finding {kind} repos ...")
            image = Image()
            location = f"{image.raspberry_images[kind]}"
            repos = [location]

            latest = {
                'date': "1900-01-01"
            }
            for repo in repos:
                versions, downloads = Image.versions(repo)
                for version, download in zip(versions, downloads):
                    entry = {
                        "version": version,
                        "tag": version.replace("raspios_", "").replace("_armhf", ""),
                        "url": download,
                        "date": version.split("-", 1)[1],
                        "type": kind,
                        "os": "raspberryos",
                    }
                    data[kind].append(entry)
                    if entry["date"] >= latest['date']:
                        latest = dict(entry)
                        latest["tag"] = f"latest-{kind}"

                data[kind].append(latest)

        if refresh or not cache.exists():
            os.system("mkdir -p ~/.cloudmesh/cmburn")
            fetch_kind(kind="lite")
            fetch_kind(kind="full")
            writefile(cache, yaml.dump(data))

        data = readfile(cache)
        data = yaml.safe_load(readfile(cache))
        # convert to array
        result = data["lite"] + data["full"] + Ubuntu.distribution

        return result

    @staticmethod
    def versions(repo=None):
        """
        Fetch and list available image versions and their download URLs
        """
        result = requests.get(repo, verify=False)
        lines = result.text.split(' ')
        d = []
        v = []
        for line in lines:
            if 'href="' in line and "</td>" in line:
                line = line.split('href="')[1]
                line = line.split('/')[0]
                v.append(line)
                download = Image.find_image_zip(repo, line)
                d.append(download)
        return v, d

    @staticmethod
    def find_image_zip(repo=None, version=None):
        url = f"{repo}/{version}/"

        result = requests.get(url, verify=False)
        lines = result.text.split(' ')
        for line in lines:
            if '.zip"' in line and "</td>" in line:
                line = line.split('href="')[1]
                line = line.split('"')[0]
                link = f"{repo}/{version}/{line}"
                return link
        return None

    @staticmethod
    def latest_version(kind="lite"):
        data = Image().read_version_cache()

        source_url = data[kind]["latest"]['url']

        return os.path.basename(source_url)[:-4]

    @staticmethod
    def get_name(url):
        return os.path.basename(url).replace('.zip', '')

    # noinspection PyBroadException
    def fetch(self, url=None, tag=None, verify=True):
        """
        Download the image from the URL in self.image_name
        If it is 'latest', download the latest image - afterwards use
          cm-pi-burn image ls
          to get the name of the downloaded latest image.
        """

        if url is None:
            data = Image().create_version_cache(refresh=False)  # noqa: F841

            image = Image().find(tag=tag)

            if image is None:
                Console.error("No matching image found.")
                return None
            elif len(image) > 1:
                Console.error("Too many images found")
                print(Printer.write(image,
                                    order=["tag", "version"],
                                    header=["Tag", "Version"]))
                return None

            image = image[0]

            # image_path = Image().directory + "/" + Image.get_name(image[
            # "url"]) + ".img"

        if not os.path.exists(self.directory):
            os.makedirs(self.directory)
        os.chdir(self.directory)
        # get image URL metadata, including the name of the latest image after
        #   the 'latest' URL redirects to the URL of the actual image

        if "ubuntu" in image["url"]:
            source_url = requests.head(image["url"], allow_redirects=True).url
            size = requests.get(image["url"], verify=False, stream=True).headers['Content-length']

            xz_filename = os.path.basename(source_url)
            img_filename = xz_filename.replace('.xz', '')
            img_file = Path(Path(self.directory) / Path(img_filename))
            # xz_file = Path(Path(self.directory) / Path(xz_filename))

            print(f"Downloading {xz_filename}")

            os.chdir(path_expand("~/.cloudmesh/cmburn/images"))

            if os.path.exists(img_filename):
                print()
                Console.warning(f"The file is already downloaded. Found at:\n\n"
                                f"    {img_file}\n")
                return img_file
            os.system(f'wget -O {xz_filename} {image["url"]}')

            print(f"Extracting {img_filename}")
            self.unzip_image(xz_filename)
            try:
                Path(xz_filename).unlink()
            except:  # noqa: E722
                pass
            return img_filename

        else:
            source_url = requests.head(image["url"], allow_redirects=True).url
            size = requests.get(image["url"], verify=False, stream=True).headers['Content-length']
            zip_filename = os.path.basename(source_url)
            img_filename = zip_filename.replace('.zip', '.img')
            sha1_filename = zip_filename + '.sha1'
            sha256_filename = zip_filename + '.sha256'

            print(f"Downloading {zip_filename}")

            img_file = Path(Path(self.directory) / Path(img_filename))
            zip_file = Path(Path(self.directory) / Path(zip_filename))
            # sha1_file = Path(Path(self.directory) / Path(sha1_filename))
            # sha256_file = Path(Path(self.directory) / Path(sha256_filename))

            # cancel if image already downloaded
            if os.path.exists(img_filename):
                print()
                Console.warning(f"The file is already downloaded. Found at:\n\n"
                                f"    {img_file}\n")
                return img_file

            # download the image, unzip it, and delete the zip file
            image['sha1'] = image['url'] + ".sha1"
            image['sha256'] = image['url'] + ".sha256"
            if verify:
                os.system(f'wget -O {sha1_filename} {image["sha1"]}')
                os.system(f'wget -O {sha256_filename} {image["sha256"]}')

            os.system(f'wget -O {zip_filename} {image["url"]}')

            if verify:
                sha1 = sha1sum(zip_file)
                sha256 = sha256sum(zip_filename)

                f_sha1 = readfile(sha1_filename).split(" ")[0]
                f_sha256 = readfile(sha256_filename).split(" ")[0]

                if f_sha1 == sha1:
                    Console.ok("SHA1 is ok")
                if f_sha256 == sha256:
                    Console.ok("SHA256 is ok")

            zip_size = os.path.getsize(zip_file)
            if int(size) != zip_size:
                Console.error(f"Repository reported zip size {size} does not equal "
                              f"download size {zip_size}. Please try again.")
                return None

            #   if latest:  # rename filename from 'latest' to the actual image name
            #        Path('raspbian_lite_latest').rename(zip_filename)

            print(f"Extracting {img_filename}")
            self.unzip_image(zip_filename)
            Path(zip_filename).unlink()
            return img_filename

    def unzip_image(self, zip_filename=None):
        """
        Unzip image.zip to image.img
        """
        image = Image()
        os.chdir(image.directory)
        # img_filename = zip_filename.replace('.zip', '.img')
        if ".zip" in zip_filename:
            zipfile.ZipFile(zip_filename).extractall()
        elif ".xz" in zip_filename or ".gz" in zip_filename:

            os.system(f"unxz {zip_filename}")
            # with tarfile.open(zip_filename) as f:
            #   content = f.extractall()

        else:
            raise ValueError("unkown zip format")

    def verify(self):
        """
        verify if the image is ok, use SHA
        """
        raise NotImplementedError

    # noinspection PyBroadException
    def rm(self, image="lite"):
        """
        Delete a downloaded image
        """
        # tag = tag or "latest"
        # if tag == "latest":
        #   tag = latest_version(kind="lite")

        for ending in [".img", ".zip"]:
            try:
                Path(Path(self.directory) / Path(image + ending)).unlink()
            except Exception as e:  # noqa: F841
                pass

    def ls(self):
        """
        List all downloaded images
        """
        images_dir = Path(self.directory)
        images = [str(x).replace(self.directory + '/', '').replace('.img', '')
                  for x in images_dir.glob('*.img')]

        print('Available images in', self.directory)
        # print(columns * '=')

        banner('Available Images')

        print(textwrap.indent('\n'.join(images), prefix="    * "))
        print()



import os
from cloudmesh.common.Shell import Shell

from cloudmesh.burn.image import Image
from cloudmesh.burn.util import os_is_linux
from cloudmesh.burn.util import os_is_pi
from cloudmesh.burn.util import os_is_mac
from cloudmesh.burn.util import os_is_windows
from cloudmesh.common.console import Console
from cloudmesh.common.sudo import Sudo


class Imager:

    @staticmethod
    def installed():
        r = Shell.which("rpi-imager")
        return r is not None

    @staticmethod
    def install(force=False):
        if os_is_mac():
            return
        if not Imager.installed() or force:
            if os_is_linux() or os_is_pi():
                Sudo.password()
                os.system("sudo apt uninstall -y rpi-imager")
            else:
                Console.warning("Installation is not supported")

    @staticmethod
    def fetch(tag=None):
        tag = tag or ['latest-lite']
        Image.create_version_cache()
        file = Image().fetch(tag=tag)

        return file

    @staticmethod
    def launch(file=None):

        if file is not None:

            if not str(file).endswith(".img"):
                raise ValueError(f"file {file} does not end with .img")

            if not os.path.exists(file):
                raise ValueError(f"image file {file} does not exist")

        elif file is None:
            file = ""

        Imager.install()

        if os_is_linux() or os_is_pi():
            Sudo.password()
            os.system(f"sudo rpi-imager {file}")
        elif os_is_mac():
            os.system(f"/Applications/Raspberry\ Pi\ Imager.app/Contents/MacOS/rpi-imager {file} "  # noqa: W605
                      "> /dev/null 2>&1")  # noqa: W605
        elif os_is_windows():
            os.chdir(r"C:\Program Files (x86)\Raspberry Pi Imager")
            os.system("rpi-imager.exe")



#
# python cloudmesh/burn/raspberryos/test.py
#
from cloudmesh.burn.raspberryos.cmdline import Cmdline
from cloudmesh.burn.raspberryos.runfirst import Runfirst
from cloudmesh.common.parameter import Parameter

cmsline = Cmdline()
# cmsline.read(filename='/Volumes/boot/cmdline.txt')
# cmsline.write(filename="test.txt")

print(cmsline.get())
print()


names = Parameter.expand("red,red[0-1]")
ips = Parameter.expand("10.0.0.[1-3]")

runfirst = Runfirst()
runfirst.set_key()
runfirst.set_wifi("SSID", "PASSWORD")
runfirst.set_locale()
runfirst.set_hostname("red2")
runfirst.set_hosts(names, ips)
runfirst.set_static_ip(ip="10.0.0.4")
runfirst.set_password(password="password")
runfirst.info()

print(runfirst.get())



import os

from cloudmesh.burn.burner.raspberryos import Burner as RaspberryOsBurner
from cloudmesh.burn.usb import USB
from cloudmesh.burn.util import os_is_linux
from cloudmesh.burn.util import os_is_mac
from cloudmesh.burn.util import os_is_pi
from cloudmesh.common.JobScript import JobScript
from cloudmesh.common.Tabulate import Printer
from cloudmesh.common.console import Console
from cloudmesh.common.util import banner
from cloudmesh.common.util import path_expand


# class Burner(AbstractBurner):
class Burner:

    def __init__(self, card_os="raspberryos"):
        if "raspberry" in card_os:
            self.burner = RaspberryOsBurner()
        else:
            Console.error("Card OS not supported")

    def get(self):
        return self.burner

    @staticmethod
    def detect():
        if os_is_mac():
            details = USB.get_from_diskutil()
        else:
            details = USB.get_from_dmesg()
        return details

    def shrink(self, image=None):
        if image is None:
            Console.error("Image must have a value")
        image = path_expand(image)
        command = f"sudo /usr/local/bin/pishrink.sh {image}"
        print(command)
        os.system(command)

    def install(self):
        """
        Installs /usr/local/bin/pishrink.sh
        Installs parted
        :return:
        :rtype:
        """

        if os_is_mac():
            Console.error("This command is not supported on MacOS")
            return ""
        else:
            banner("Installing pishrink.sh into /usr/local/bin")
            script = \
                """
                wget https://raw.githubusercontent.com/Drewsif/PiShrink/master/pishrink.sh
                chmod +x pishrink.sh
                sudo mv pishrink.sh /usr/local/bin
                """

            result = JobScript.execute(script)
            print(Printer.write(result,
                                order=["name", "command", "status", "stdout", "returncode"]))

            if os_is_linux() or os_is_pi():
                banner("Installing pishrink.sh into /usr/local/bin")
                script = \
                    """
                    sudo apt install parted -y > $HOME/tmp.log
                    """

                result = JobScript.execute(script)
                print(Printer.write(result,
                                    order=["name", "command", "status", "stdout", "returncode"]))

    def firmware(self, action="check"):
        self.burner.firmware(action=action)

    def check(self, device=None):
        self.burner.check(device=device)

    def configure_wifi(self, ssid, psk=None, country=None, host=None):
        self.burner.configure_wifi(ssid, psk=psk, country=country, host=host)

    def mac(self, hostnames=None):
        self.burner.mac(hostnames=hostnames)

    def set_hostname(self, hostname):
        self.burner.set_hostname(hostname)

    def set_static_ip(self, ip):
        self.burner.set_static_ip(ip)

    def set_cmdline(self, cmdline):
        self.burner.set_cmdline(cmdline)

    def set_key(self, key):
        self.burner.set_key(key)

    def keyboard(self, country=None):
        self.burner.keyboard(country=country)

    def enable_ssh(self):
        self.burner.enable_ssh()

    def cluster(self, arguments=None):
        self.burner.cluster(arguments=arguments)


# multi_self.burner.burn_inventory



import textwrap
from cloudmesh.common.util import path_expand, readfile, writefile
from cloudmesh.common.console import Console
from cloudmesh.common.Shell import Shell
from passlib.hash import sha256_crypt


def dedent(content):
    return textwrap.dedent(content).strip()


class Runfirst:
    """
    A builder-like class for constructing a firstrun.sh script to be run
    at boot by a Raspberry Pi
    """

    SCRIPT_NAME = 'firstrun.sh'

    def __init__(self):
        self.key = None
        self.hostname = None
        self.timezone = None
        self.ssid = None
        self.wifipasswd = None
        self.locale = None
        self.country = None
        self.script = None
        self.etc_hosts = None
        self.static_ip_info = None
        self.password = None
        self.bridge = None

    def info(self):
        print("Key:     ", self.key[0:20], "...", self.key[-20:].strip())
        print("Hostname:", self.hostname)
        print("Timezone:", self.timezone)
        print("SSID:    ", self.ssid)
        print("Locale:  ", self.locale)
        print("Country: ", self.country)
        print()

    def set_key(self, key=None):
        if key is None:
            self.key = readfile("~/.ssh/id_rsa.pub").strip()
        else:
            self.key = key

    def enable_bridge(self):
        """
        Enables the wifi bridge
        """
        self.bridge = True

    def set_hostname(self, name):
        #
        # I do not yet know how they write the hostname ... into /etc/hostname
        #
        self.hostname = name

    def set_hosts(self, names, ips):
        """
        Sets the /etc/hosts file information

        :param names: list of names
        :type names: list
        :param ips: list of ips
        :type ips: list
        """
        self.etc_hosts = dict(zip(names, ips))

    def set_wifi(self, ssid, passwd, country="US"):
        """
        sets the wifi password

        :param ssid:
        :type ssid:
        :param passwd:
        :type passwd:
        :return:
        :rtype:
        """
        # Same as ubuntu/ previous raspberry, write to boot partition and then
        # just add to first run to place it in place
        self.ssid = ssid
        self.wifipasswd = passwd
        self.country = country

    def set_locale(self, timezone=None, locale=None):
        """
        sets the locale of the OS

        :return:
        :rtype:
        """
        # same as ubuntu or previous raspberry os. E.g. get form current
        # machine and create the KBEOF
        # see the get script how they include it.
        #
        if not timezone:
            self.timezone = "America/Indiana/Indianapolis"
        else:
            self.timezone = timezone
        if not locale:
            self.locale = "us"
        else:
            self.locale = locale

    def set_static_ip(self, interface='eth0', ip=None, subnet_mask='24', router=None, dns=None):
        """
        Sets a static IP on the specified interface
        """
        if ip is None:
            raise Exception("Missing ip arg. None supplied")

        self.static_ip_info = [interface, ip, subnet_mask, router, dns]

    def set_password(self, password=None):
        if password is None:
            raise Exception("Missing password arg. None supplied")

        self.password = password

    def _get_bridge_script(self):
        """
        If self.bridge is True, then enable a bridge from eth0 to wlan0
        """
        if self.bridge:
            script = []
            script += ["sudo sed -i 's/#net\\.ipv4\\.ip_forward=1/net.ipv4.ip_forward=1/' /etc/sysctl.conf"]
            script += ["sudo iptables -A FORWARD -i eth0 -o wlan0 -j ACCEPT"]
            script += ["sudo iptables -A FORWARD -i wlan0 -o eth0 -m state --state ESTABLISHED,RELATED -j ACCEPT"]
            script += ["sudo iptables -t nat -A POSTROUTING -o wlan0 -j MASQUERADE"]
            script += ['sh -c "iptables-save > /etc/iptables.ipv4.nat"']
            script += ["sudo sed -i '$i iptables-restore < /etc/iptables.ipv4.nat' /etc/rc.local"]
            return '\n'.join(script)
        else:
            return ""

    def _get_password_script(self):
        #
        # BUG: hash should be hash_str, you can not use hash and .hash as function
        #
        script = []
        if not self.password:
            return ""
        # repeatable salt if needed for testing
        # hash = sha256_crypt.using(salt='qY2oeR.YpL', rounds=5000).hash(
        # self.password)
        hash = sha256_crypt.using(rounds=5000).hash(self.password)
        script.append(f'echo "$FIRSTUSER:"\'{hash}\' | chpasswd -e')
        return '\n'.join(script)

    def _get_static_ip_script(self):
        """
        If the self.interface_ip pair is not None, then return the script
        to configure it
        """
        script = []
        if not self.static_ip_info:
            return ""
        interface, ip, mask, router, dns = self.static_ip_info
        script.append(f'echo "interface {interface}" >> /etc/dhcpcd.conf')
        script.append(f'echo "static ip_address={ip}/{mask}" >> /etc/dhcpcd.conf')
        if router:
            script.append(f'echo "static routers={router}" >> /etc/dhcpcd.conf')
        if dns:
            dns = ' '.join(dns)
            script.append(f'echo "static domain_name_servers={dns}" >> /etc/dhcpcd.conf')
        return '\n'.join(script)

    def _get_etc_hosts_script(self):
        """
        If self.etc_hosts is not None, then we must append the known hosts
        to /etc/hosts
        """
        script = []
        if self.etc_hosts:
            for hostname, ip in self.etc_hosts.items():
                script.append(f"echo {ip}\t{hostname} >> /etc/hosts")
        return '\n'.join(script)

    def _get_wifi_config(self):
        if self.ssid:
            script = f"""
                cat >/etc/wpa_supplicant/wpa_supplicant.conf <<WPAEOF
                country={self.country}
                ctrl_interface=DIR=/var/run/wpa_supplicant GROUP=netdev
                ap_scan=1
                update_config=1
                network={{
                    ssid="{self.ssid}"
                    psk="{self.wifipasswd}"
                }}
                WPAEOF
                chmod 600 /etc/wpa_supplicant/wpa_supplicant.conf
                rfkill unblock wifi
                for filename in /var/lib/systemd/rfkill/*:wlan ; do
                    echo 0 > $filename
                done"""
            return dedent(script)
        else:
            return ""

    def _writefile(self, name, content):
        data = dedent(f"""
        cat > {name} <<EOF
        {content}
        EOF
        """)
        return data

    def get_manager(self):
        return self.get()

    def get_worker(self):
        return self.get()

    def write(self, filename=None):
        """
        Write the cmdline config to the specified filename
        """
        if filename is None:
            raise Exception("write called with no filename")
        if self.script is None:
            raise Exception("no script found. Did you run .get() first?")

        tmp_location = path_expand('~/.cloudmesh/cmburn/firstrun.sh.tmp')
        writefile(tmp_location, self.script)
        Shell.run(f'cat {tmp_location} | sudo tee {filename}')
        Shell.execute('rm', arguments=[tmp_location])

    def get(self, verbose=False):
        # NEEDS TO BE INDENTED THIS WAY
        # OR ELSE WRITTEN SCRIPT WILL NOT WORK
        self.script = dedent(f'''
#!/bin/bash
set +e
CURRENT_HOSTNAME=`cat /etc/hostname | tr -d " \\t\\n\\r"`
echo {self.hostname} >/etc/hostname
sed -i "s/127\\.0\\.1\\.1.*$CURRENT_HOSTNAME/127.0.1.1\\t{self.hostname}/g" /etc/hosts
{self._get_etc_hosts_script()}
{self._get_static_ip_script()}
FIRSTUSER=`getent passwd 1000 | cut -d: -f1`
FIRSTUSERHOME=`getent passwd 1000 | cut -d: -f6`
install -o "$FIRSTUSER" -m 700 -d "$FIRSTUSERHOME/.ssh"
install -o "$FIRSTUSER" -m 600 <(echo "{self.key}") "$FIRSTUSERHOME/.ssh/authorized_keys"
cp /etc/ssh/sshd_config /etc/ssh/sshd_config.orig
echo 'PasswordAuthentication no' >>/etc/ssh/sshd_config
systemctl enable ssh
{self._get_bridge_script()}
{self._get_password_script()}
{self._get_wifi_config()}
rm -f /etc/xdg/autostart/piwiz.desktop
rm -f /etc/localtime
echo "{self.timezone}" >/etc/timezone
dpkg-reconfigure -f noninteractive tzdata
cat >/etc/default/keyboard <<KBEOF
XKBMODEL="pc105"
XKBLAYOUT="{self.locale}"
XKBVARIANT=""
XKBOPTIONS=""
KBEOF
dpkg-reconfigure -f noninteractive keyboard-configuration
rm -f /boot/{Runfirst.SCRIPT_NAME}
sed -i 's| systemd.run.*||g' /boot/cmdline.txt
exit 0
        ''')

        if verbose:
            Console.info(f'{Runfirst.SCRIPT_NAME}.sh for {self.hostname}')
            Console.info(self.script)
        return self.script



import textwrap


# Locale.gen
class Locale:

    @property
    def gen(self):

        data = textwrap.dedent("""
        # This file lists locales that you wish to have built. You can find a list
        # of valid supported locales at /usr/share/i18n/SUPPORTED, and you can add
        # user defined locales to /usr/local/share/i18n/SUPPORTED. If you change
        # this file, you need to rerun locale-gen.
        #
        # aa_DJ ISO-8859-1
        # aa_DJ.UTF-8 UTF-8
        # aa_ER UTF-8
        # aa_ER@saaho UTF-8
        # aa_ET UTF-8
        # af_ZA ISO-8859-1
        # af_ZA.UTF-8 UTF-8
        # agr_PE UTF-8
        # ak_GH UTF-8
        # am_ET UTF-8
        # an_ES ISO-8859-15
        # an_ES.UTF-8 UTF-8
        # anp_IN UTF-8
        # ar_AE ISO-8859-6
        # ar_AE.UTF-8 UTF-8
        # ar_BH ISO-8859-6
        # ar_BH.UTF-8 UTF-8
        # ar_DZ ISO-8859-6
        # ar_DZ.UTF-8 UTF-8
        # ar_EG ISO-8859-6
        # ar_EG.UTF-8 UTF-8
        # ar_IN UTF-8
        # ar_IQ ISO-8859-6
        # ar_IQ.UTF-8 UTF-8
        # ar_JO ISO-8859-6
        # ar_JO.UTF-8 UTF-8
        # ar_KW ISO-8859-6
        # ar_KW.UTF-8 UTF-8
        # ar_LB ISO-8859-6
        # ar_LB.UTF-8 UTF-8
        # ar_LY ISO-8859-6
        # ar_LY.UTF-8 UTF-8
        # ar_MA ISO-8859-6
        # ar_MA.UTF-8 UTF-8
        # ar_OM ISO-8859-6
        # ar_OM.UTF-8 UTF-8
        # ar_QA ISO-8859-6
        # ar_QA.UTF-8 UTF-8
        # ar_SA ISO-8859-6
        # ar_SA.UTF-8 UTF-8
        # ar_SD ISO-8859-6
        # ar_SD.UTF-8 UTF-8
        # ar_SS UTF-8
        # ar_SY ISO-8859-6
        # ar_SY.UTF-8 UTF-8
        # ar_TN ISO-8859-6
        # ar_TN.UTF-8 UTF-8
        # ar_YE ISO-8859-6
        # ar_YE.UTF-8 UTF-8
        # as_IN UTF-8
        # ast_ES ISO-8859-15
        # ast_ES.UTF-8 UTF-8
        # ayc_PE UTF-8
        # az_AZ UTF-8
        # az_IR UTF-8
        # be_BY CP1251
        # be_BY.UTF-8 UTF-8
        # be_BY@latin UTF-8
        # bem_ZM UTF-8
        # ber_DZ UTF-8
        # ber_MA UTF-8
        # bg_BG CP1251
        # bg_BG.UTF-8 UTF-8
        # bhb_IN.UTF-8 UTF-8
        # bho_IN UTF-8
        # bho_NP UTF-8
        # bi_VU UTF-8
        # bn_BD UTF-8
        # bn_IN UTF-8
        # bo_CN UTF-8
        # bo_IN UTF-8
        # br_FR ISO-8859-1
        # br_FR.UTF-8 UTF-8
        # br_FR@euro ISO-8859-15
        # brx_IN UTF-8
        # bs_BA ISO-8859-2
        # bs_BA.UTF-8 UTF-8
        # byn_ER UTF-8
        # ca_AD ISO-8859-15
        # ca_AD.UTF-8 UTF-8
        # ca_ES ISO-8859-1
        # ca_ES.UTF-8 UTF-8
        # ca_ES@euro ISO-8859-15
        # ca_ES@valencia UTF-8
        # ca_FR ISO-8859-15
        # ca_FR.UTF-8 UTF-8
        # ca_IT ISO-8859-15
        # ca_IT.UTF-8 UTF-8
        # ce_RU UTF-8
        # chr_US UTF-8
        # cmn_TW UTF-8
        # crh_UA UTF-8
        # cs_CZ ISO-8859-2
        # cs_CZ.UTF-8 UTF-8
        # csb_PL UTF-8
        # cv_RU UTF-8
        # cy_GB ISO-8859-14
        # cy_GB.UTF-8 UTF-8
        # da_DK ISO-8859-1
        # da_DK.UTF-8 UTF-8
        # de_AT ISO-8859-1
        # de_AT.UTF-8 UTF-8
        # de_AT@euro ISO-8859-15
        # de_BE ISO-8859-1
        # de_BE.UTF-8 UTF-8
        # de_BE@euro ISO-8859-15
        # de_CH ISO-8859-1
        # de_CH.UTF-8 UTF-8
        # de_DE ISO-8859-1
        # de_DE.UTF-8 UTF-8
        # de_DE@euro ISO-8859-15
        # de_IT ISO-8859-1
        # de_IT.UTF-8 UTF-8
        # de_LI.UTF-8 UTF-8
        # de_LU ISO-8859-1
        # de_LU.UTF-8 UTF-8
        # de_LU@euro ISO-8859-15
        # doi_IN UTF-8
        # dsb_DE UTF-8
        # dv_MV UTF-8
        # dz_BT UTF-8
        # el_CY ISO-8859-7
        # el_CY.UTF-8 UTF-8
        # el_GR ISO-8859-7
        # el_GR.UTF-8 UTF-8
        # el_GR@euro ISO-8859-7
        # en_AG UTF-8
        # en_AU ISO-8859-1
        # en_AU.UTF-8 UTF-8
        # en_BW ISO-8859-1
        # en_BW.UTF-8 UTF-8
        # en_CA ISO-8859-1
        # en_CA.UTF-8 UTF-8
        # en_DK ISO-8859-1
        # en_DK.ISO-8859-15 ISO-8859-15
        # en_DK.UTF-8 UTF-8
        # en_GB ISO-8859-1
        # en_GB.ISO-8859-15 ISO-8859-15
        en_GB.UTF-8 UTF-8
        # en_HK ISO-8859-1
        # en_HK.UTF-8 UTF-8
        # en_IE ISO-8859-1
        # en_IE.UTF-8 UTF-8
        # en_IE@euro ISO-8859-15
        # en_IL UTF-8
        # en_IN UTF-8
        # en_NG UTF-8
        # en_NZ ISO-8859-1
        # en_NZ.UTF-8 UTF-8
        # en_PH ISO-8859-1
        # en_PH.UTF-8 UTF-8
        # en_SC.UTF-8 UTF-8
        # en_SG ISO-8859-1
        # en_SG.UTF-8 UTF-8
        # en_US ISO-8859-1
        # en_US.ISO-8859-15 ISO-8859-15
        # en_US.UTF-8 UTF-8
        # en_ZA ISO-8859-1
        # en_ZA.UTF-8 UTF-8
        # en_ZM UTF-8
        # en_ZW ISO-8859-1
        # en_ZW.UTF-8 UTF-8
        # eo UTF-8
        # es_AR ISO-8859-1
        # es_AR.UTF-8 UTF-8
        # es_BO ISO-8859-1
        # es_BO.UTF-8 UTF-8
        # es_CL ISO-8859-1
        # es_CL.UTF-8 UTF-8
        # es_CO ISO-8859-1
        # es_CO.UTF-8 UTF-8
        # es_CR ISO-8859-1
        # es_CR.UTF-8 UTF-8
        # es_CU UTF-8
        # es_DO ISO-8859-1
        # es_DO.UTF-8 UTF-8
        # es_EC ISO-8859-1
        # es_EC.UTF-8 UTF-8
        # es_ES ISO-8859-1
        # es_ES.UTF-8 UTF-8
        # es_ES@euro ISO-8859-15
        # es_GT ISO-8859-1
        # es_GT.UTF-8 UTF-8
        # es_HN ISO-8859-1
        # es_HN.UTF-8 UTF-8
        # es_MX ISO-8859-1
        # es_MX.UTF-8 UTF-8
        # es_NI ISO-8859-1
        # es_NI.UTF-8 UTF-8
        # es_PA ISO-8859-1
        # es_PA.UTF-8 UTF-8
        # es_PE ISO-8859-1
        # es_PE.UTF-8 UTF-8
        # es_PR ISO-8859-1
        # es_PR.UTF-8 UTF-8
        # es_PY ISO-8859-1
        # es_PY.UTF-8 UTF-8
        # es_SV ISO-8859-1
        # es_SV.UTF-8 UTF-8
        # es_US ISO-8859-1
        # es_US.UTF-8 UTF-8
        # es_UY ISO-8859-1
        # es_UY.UTF-8 UTF-8
        # es_VE ISO-8859-1
        # es_VE.UTF-8 UTF-8
        # et_EE ISO-8859-1
        # et_EE.ISO-8859-15 ISO-8859-15
        # et_EE.UTF-8 UTF-8
        # eu_ES ISO-8859-1
        # eu_ES.UTF-8 UTF-8
        # eu_ES@euro ISO-8859-15
        # eu_FR ISO-8859-1
        # eu_FR.UTF-8 UTF-8
        # eu_FR@euro ISO-8859-15
        # fa_IR UTF-8
        # ff_SN UTF-8
        # fi_FI ISO-8859-1
        # fi_FI.UTF-8 UTF-8
        # fi_FI@euro ISO-8859-15
        # fil_PH UTF-8
        # fo_FO ISO-8859-1
        # fo_FO.UTF-8 UTF-8
        # fr_BE ISO-8859-1
        # fr_BE.UTF-8 UTF-8
        # fr_BE@euro ISO-8859-15
        # fr_CA ISO-8859-1
        # fr_CA.UTF-8 UTF-8
        # fr_CH ISO-8859-1
        # fr_CH.UTF-8 UTF-8
        # fr_FR ISO-8859-1
        # fr_FR.UTF-8 UTF-8
        # fr_FR@euro ISO-8859-15
        # fr_LU ISO-8859-1
        # fr_LU.UTF-8 UTF-8
        # fr_LU@euro ISO-8859-15
        # fur_IT UTF-8
        # fy_DE UTF-8
        # fy_NL UTF-8
        # ga_IE ISO-8859-1
        # ga_IE.UTF-8 UTF-8
        # ga_IE@euro ISO-8859-15
        # gd_GB ISO-8859-15
        # gd_GB.UTF-8 UTF-8
        # gez_ER UTF-8
        # gez_ER@abegede UTF-8
        # gez_ET UTF-8
        # gez_ET@abegede UTF-8
        # gl_ES ISO-8859-1
        # gl_ES.UTF-8 UTF-8
        # gl_ES@euro ISO-8859-15
        # gu_IN UTF-8
        # gv_GB ISO-8859-1
        # gv_GB.UTF-8 UTF-8
        # ha_NG UTF-8
        # hak_TW UTF-8
        # he_IL ISO-8859-8
        # he_IL.UTF-8 UTF-8
        # hi_IN UTF-8
        # hif_FJ UTF-8
        # hne_IN UTF-8
        # hr_HR ISO-8859-2
        # hr_HR.UTF-8 UTF-8
        # hsb_DE ISO-8859-2
        # hsb_DE.UTF-8 UTF-8
        # ht_HT UTF-8
        # hu_HU ISO-8859-2
        # hu_HU.UTF-8 UTF-8
        # hy_AM UTF-8
        # hy_AM.ARMSCII-8 ARMSCII-8
        # ia_FR UTF-8
        # id_ID ISO-8859-1
        # id_ID.UTF-8 UTF-8
        # ig_NG UTF-8
        # ik_CA UTF-8
        # is_IS ISO-8859-1
        # is_IS.UTF-8 UTF-8
        # it_CH ISO-8859-1
        # it_CH.UTF-8 UTF-8
        # it_IT ISO-8859-1
        # it_IT.UTF-8 UTF-8
        # it_IT@euro ISO-8859-15
        # iu_CA UTF-8
        # ja_JP.EUC-JP EUC-JP
        # ja_JP.UTF-8 UTF-8
        # ka_GE GEORGIAN-PS
        # ka_GE.UTF-8 UTF-8
        # kab_DZ UTF-8
        # kk_KZ PT154
        # kk_KZ.RK1048 RK1048
        # kk_KZ.UTF-8 UTF-8
        # kl_GL ISO-8859-1
        # kl_GL.UTF-8 UTF-8
        # km_KH UTF-8
        # kn_IN UTF-8
        # ko_KR.EUC-KR EUC-KR
        # ko_KR.UTF-8 UTF-8
        # kok_IN UTF-8
        # ks_IN UTF-8
        # ks_IN@devanagari UTF-8
        # ku_TR ISO-8859-9
        # ku_TR.UTF-8 UTF-8
        # kw_GB ISO-8859-1
        # kw_GB.UTF-8 UTF-8
        # ky_KG UTF-8
        # lb_LU UTF-8
        # lg_UG ISO-8859-10
        # lg_UG.UTF-8 UTF-8
        # li_BE UTF-8
        # li_NL UTF-8
        # lij_IT UTF-8
        # ln_CD UTF-8
        # lo_LA UTF-8
        # lt_LT ISO-8859-13
        # lt_LT.UTF-8 UTF-8
        # lv_LV ISO-8859-13
        # lv_LV.UTF-8 UTF-8
        # lzh_TW UTF-8
        # mag_IN UTF-8
        # mai_IN UTF-8
        # mai_NP UTF-8
        # mfe_MU UTF-8
        # mg_MG ISO-8859-15
        # mg_MG.UTF-8 UTF-8
        # mhr_RU UTF-8
        # mi_NZ ISO-8859-13
        # mi_NZ.UTF-8 UTF-8
        # miq_NI UTF-8
        # mjw_IN UTF-8
        # mk_MK ISO-8859-5
        # mk_MK.UTF-8 UTF-8
        # ml_IN UTF-8
        # mn_MN UTF-8
        # mni_IN UTF-8
        # mr_IN UTF-8
        # ms_MY ISO-8859-1
        # ms_MY.UTF-8 UTF-8
        # mt_MT ISO-8859-3
        # mt_MT.UTF-8 UTF-8
        # my_MM UTF-8
        # nan_TW UTF-8
        # nan_TW@latin UTF-8
        # nb_NO ISO-8859-1
        # nb_NO.UTF-8 UTF-8
        # nds_DE UTF-8
        # nds_NL UTF-8
        # ne_NP UTF-8
        # nhn_MX UTF-8
        # niu_NU UTF-8
        # niu_NZ UTF-8
        # nl_AW UTF-8
        # nl_BE ISO-8859-1
        # nl_BE.UTF-8 UTF-8
        # nl_BE@euro ISO-8859-15
        # nl_NL ISO-8859-1
        # nl_NL.UTF-8 UTF-8
        # nl_NL@euro ISO-8859-15
        # nn_NO ISO-8859-1
        # nn_NO.UTF-8 UTF-8
        # nr_ZA UTF-8
        # nso_ZA UTF-8
        # oc_FR ISO-8859-1
        # oc_FR.UTF-8 UTF-8
        # om_ET UTF-8
        # om_KE ISO-8859-1
        # om_KE.UTF-8 UTF-8
        # or_IN UTF-8
        # os_RU UTF-8
        # pa_IN UTF-8
        # pa_PK UTF-8
        # pap_AW UTF-8
        # pap_CW UTF-8
        # pl_PL ISO-8859-2
        # pl_PL.UTF-8 UTF-8
        # ps_AF UTF-8
        # pt_BR ISO-8859-1
        # pt_BR.UTF-8 UTF-8
        # pt_PT ISO-8859-1
        # pt_PT.UTF-8 UTF-8
        # pt_PT@euro ISO-8859-15
        # quz_PE UTF-8
        # raj_IN UTF-8
        # ro_RO ISO-8859-2
        # ro_RO.UTF-8 UTF-8
        # ru_RU ISO-8859-5
        # ru_RU.CP1251 CP1251
        # ru_RU.KOI8-R KOI8-R
        # ru_RU.UTF-8 UTF-8
        # ru_UA KOI8-U
        # ru_UA.UTF-8 UTF-8
        # rw_RW UTF-8
        # sa_IN UTF-8
        # sah_RU UTF-8
        # sat_IN UTF-8
        # sc_IT UTF-8
        # sd_IN UTF-8
        # sd_IN@devanagari UTF-8
        # se_NO UTF-8
        # sgs_LT UTF-8
        # shn_MM UTF-8
        # shs_CA UTF-8
        # si_LK UTF-8
        # sid_ET UTF-8
        # sk_SK ISO-8859-2
        # sk_SK.UTF-8 UTF-8
        # sl_SI ISO-8859-2
        # sl_SI.UTF-8 UTF-8
        # sm_WS UTF-8
        # so_DJ ISO-8859-1
        # so_DJ.UTF-8 UTF-8
        # so_ET UTF-8
        # so_KE ISO-8859-1
        # so_KE.UTF-8 UTF-8
        # so_SO ISO-8859-1
        # so_SO.UTF-8 UTF-8
        # sq_AL ISO-8859-1
        # sq_AL.UTF-8 UTF-8
        # sq_MK UTF-8
        # sr_ME UTF-8
        # sr_RS UTF-8
        # sr_RS@latin UTF-8
        # ss_ZA UTF-8
        # st_ZA ISO-8859-1
        # st_ZA.UTF-8 UTF-8
        # sv_FI ISO-8859-1
        # sv_FI.UTF-8 UTF-8
        # sv_FI@euro ISO-8859-15
        # sv_SE ISO-8859-1
        # sv_SE.ISO-8859-15 ISO-8859-15
        # sv_SE.UTF-8 UTF-8
        # sw_KE UTF-8
        # sw_TZ UTF-8
        # szl_PL UTF-8
        # ta_IN UTF-8
        # ta_LK UTF-8
        # tcy_IN.UTF-8 UTF-8
        # te_IN UTF-8
        # tg_TJ KOI8-T
        # tg_TJ.UTF-8 UTF-8
        # th_TH TIS-620
        # th_TH.UTF-8 UTF-8
        # the_NP UTF-8
        # ti_ER UTF-8
        # ti_ET UTF-8
        # tig_ER UTF-8
        # tk_TM UTF-8
        # tl_PH ISO-8859-1
        # tl_PH.UTF-8 UTF-8
        # tn_ZA UTF-8
        # to_TO UTF-8
        # tpi_PG UTF-8
        # tr_CY ISO-8859-9
        # tr_CY.UTF-8 UTF-8
        # tr_TR ISO-8859-9
        # tr_TR.UTF-8 UTF-8
        # ts_ZA UTF-8
        # tt_RU UTF-8
        # tt_RU@iqtelif UTF-8
        # ug_CN UTF-8
        # uk_UA KOI8-U
        # uk_UA.UTF-8 UTF-8
        # unm_US UTF-8
        # ur_IN UTF-8
        # ur_PK UTF-8
        # uz_UZ ISO-8859-1
        # uz_UZ.UTF-8 UTF-8
        # uz_UZ@cyrillic UTF-8
        # ve_ZA UTF-8
        # vi_VN UTF-8
        # wa_BE ISO-8859-1
        # wa_BE.UTF-8 UTF-8
        # wa_BE@euro ISO-8859-15
        # wae_CH UTF-8
        # wal_ET UTF-8
        # wo_SN UTF-8
        # xh_ZA ISO-8859-1
        # xh_ZA.UTF-8 UTF-8
        # yi_US CP1255
        # yi_US.UTF-8 UTF-8
        # yo_NG UTF-8
        # yue_HK UTF-8
        # yuw_PG UTF-8
        # zh_CN GB2312
        # zh_CN.GB18030 GB18030
        # zh_CN.GBK GBK
        # zh_CN.UTF-8 UTF-8
        # zh_HK BIG5-HKSCS
        # zh_HK.UTF-8 UTF-8
        # zh_SG GB2312
        # zh_SG.GBK GBK
        # zh_SG.UTF-8 UTF-8
        # zh_TW BIG5
        # zh_TW.EUC-TW EUC-TW
        # zh_TW.UTF-8 UTF-8
        # zu_ZA ISO-8859-1
        # zu_ZA.UTF-8 UTF-8
        """).strip()

        return data



import os

from cloudmesh.burn.image import Image


class Iso:
    distribution = {
        "latest": "https://downloads.raspberrypi.org/rpd_x86/images/rpd_x86-2021-01-12/2021-01-11-raspios-buster-i386.iso"  # noqa: E501
    }

    @staticmethod
    def get(tag="latest"):
        url = Iso.distribution[tag]
        destination = Image().directory + os.path.basename(url)

        os.system(f'wget -O {destination} {url}')



from cloudmesh.burn.ubuntu.userdata import Userdata
from cloudmesh.burn.ubuntu.networkdata import Networkdata
from cloudmesh.common.console import Console
from cloudmesh.common.util import readfile
from cloudmesh.inventory.inventory import Inventory
from cloudmesh.common.Shell import Shell
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import banner
from cloudmesh.burn.image import Image


class Configure:
    """
    This class serves to build cloud-init config files for entries in a cloudmesh inventory file.
    This accepts two params for init:
    inventory = INVENTORY_FILE
    cluster = CLUSTER_NAME

    If inventory arg is None, then default ~/.cloudmesh/inventory.yaml is used

    if cluster arg is not None, then nodes are found by searching via the
    "cluster" column in inventory

    if cluster arg is None, then nodes are found by searching for "worker" and "manager" in the
    "service" column in inventory

    Usage:
    config_generator = Configure()
    Configure.build_user_data(name=NAME) returns a Userdata builder object
    where NAME is the hostname of an entry in inventory.yaml with corresponding config options
    """
    KEY_DIR = path_expand('~/.cloudmesh/cmburn')

    def __init__(self, inventory=None, cluster=None, debug=False):
        self.debug = debug
        self.manager_public_key = None  # Populated by self.generate_ssh_key

        if inventory:
            self.inventory = Inventory(inventory)
        else:
            self.inventory = Inventory()

        if cluster is not None:
            self.nodes = self.inventory.find(cluster=cluster)
        else:
            self.nodes = self.inventory.find(service='manager') + self.inventory.find(service='worker')

        self.manager_public_key = None

        managers = self.inventory.find(service='manager')
        workers = self.inventory.find(service='worker')
        configs = managers + workers
        # Create dict for them for easy lookup
        self.configs = dict((config['host'], config) for config in configs)
        self.get_images()

    def get_images(self):
        """
        Downloads all tags found in self.configs
        """
        tags = set()
        for config in self.configs.values():
            try:
                tags.add(config['tag'])
            except KeyError as e:
                Console.warning(f'Could not find tag for {config["host"]}. Skipping')

        banner("Downloading Images", figlet=True)

        image = Image()

        for tag in tags:
            Console.info(f'Attempting to download {tag}')
            res = image.fetch(tag=[tag])
            if not res:
                Console.error('Failed Image Fetch.')
                raise Exception('Failed Image Fetch')

    def build_user_data(self, name=None, with_defaults=True, country=None,
                        add_manager_key=False, upgrade=False, with_bridge=False):
        """
        Given a name, get its config from self.inventory and create a Userdata object
        """
        if name is None:
            raise Exception('name arg supplied is None')
        elif not self.inventory.has_host(name):
            raise Exception(f'Could not find {name} in {self.inventory.filename}')
        if country is not None and len(country) != 2:
            raise Exception('Country code is not 2 characters.')

        # Get the current configurations from inventory
        hostname = self.inventory.get(name=name, attribute='host')
        keyfile = self.inventory.get(name=name, attribute='keyfile')
        if keyfile:
            keys = readfile(keyfile).strip().split('\n')
        else:
            keys = None

        if keys is None and add_manager_key:
            keys = [self.manager_public_key]
        elif add_manager_key:
            keys.append(self.manager_public_key)

        service = self.inventory.get(name=name, attribute='service')

        # Build Userdata
        user_data = Userdata()

        if with_defaults:
            user_data.with_locale().with_net_tools().with_packages(
                packages='avahi-daemon')
        if upgrade:
            user_data.with_package_update().with_package_upgrade()
        if hostname:
            user_data.with_hostname(hostname=hostname)
        if keys:
            # Disable password auth in favor of key auth
            user_data.with_ssh_password_login(ssh_pwauth=False)
            user_data.with_authorized_keys(keys=keys)
        else:
            user_data.with_default_user().with_ssh_password_login()
        # Add known hosts
        user_data.with_hosts(hosts=self.get_hosts_for(name=name))
        if country:
            user_data.with_set_wifi_country(country=country)
        if service == 'manager' and self.manager_public_key:
            # If public key is set, then we expect /boot/firmware/id_rsa and /boot/firmware/id_rsa.pub on burned card
            user_data.with_runcmd(cmd='cat /boot/firmware/id_rsa.pub > /home/ubuntu/.ssh/id_rsa.pub')\
                .with_runcmd(cmd='cat /boot/firmware/id_rsa > /home/ubuntu/.ssh/id_rsa')\
                .with_fix_user_dir_owner(user='ubuntu')\
                .with_runcmd(cmd='chmod 600 /home/ubuntu/.ssh/id_rsa')\
                .with_runcmd(cmd='sudo rm /boot/firmware/id_rsa.pub')\
                .with_runcmd(cmd='sudo rm /boot/firmware/id_rsa')
        if with_bridge:
            user_data.with_access_point_bridge()

        if self.debug:
            Console.info(f'User data for {name}:\n' + str(user_data))

        return user_data

    def build_network_data(self, name=None, ssid=None, password=None,
                           with_defaults=True):
        """
        Given a name, get its config from self.inventory and create a Networkdata object
        """
        if name is None:
            raise Exception('name arg supplied is None')
        elif not self.inventory.has_host(name):
            raise Exception(f'Could not find {name} in {self.inventory.filename}')
        if ssid or password:
            if not ssid or not password:
                raise Exception("ssid or password supplied with no corresponding ssid or password")

        # Get the current configurations from inventory
        eth0_ip = self.inventory.get(name=name, attribute='ip')
        eth0_nameservers = self.inventory.get(name=name, attribute='dns')
        eth0_gateway = self.inventory.get(name=name, attribute='router')

        network_data = Networkdata()

        if with_defaults:
            network_data.with_defaults()
        if eth0_ip:
            network_data.with_ip(ip=eth0_ip)
        if eth0_nameservers:
            network_data.with_nameservers(nameservers=eth0_nameservers)
        if eth0_gateway:
            network_data.with_gateway(gateway=eth0_gateway)
        if ssid and password:
            Console.info(f'Providing WiFi access to {name}')
            network_data.with_access_points(ssid=ssid, password=password)\
                .with_dhcp4(interfaces='wifis', interface='wlan0', dhcp4=True)\
                .with_optional(interfaces='wifis', interface='wlan0', optional=True)

        if self.debug:
            Console.info(f'Network data for {name}:\n' + str(network_data))

        return network_data

    def get_hosts_for(self, name=None):
        """
        Given a hostname, return a list of ':' separated strings of the form:

        ip:hostname

        for all hosts with ips in the inventory. Also includes mapping for own hostname in the form of 127.0.0.1:{name}

        For example, if inventory has worker001 with ip 10.1.1.1, worker002 with ip 10.1.1.2, worker003 with ip 10.1.1.3,
        then:

        self.get_hosts_for(name='worker001') returns ['127.0.0.1:worker001', '10.1.1.2:worker002', '10.1.1.3:worker003']

        Do not rely on the order of the result here
        """
        if name is None:
            raise Exception('name arg supplied is None')
        if not self.inventory.has_host(name):
            raise Exception(f'{name} could not be found in {self.inventory.filename}')

        result = [f'127.0.0.1:{name}']
        for node in self.nodes:
            if node['ip'] and name != node['host']:
                host = node['host']
                ip = node['ip']
                result += [f'{ip}:{host}']
        return result

    def generate_ssh_key(self, hostname):
        Shell.execute('mkdir', f'-p {Configure.KEY_DIR}')
        Shell.run(f'ssh-keygen -q -N "" -C "ubuntu@{hostname}" -f ' f'{Configure.KEY_DIR}/id_rsa')
        priv_key = readfile(f'{Configure.KEY_DIR}/id_rsa').strip()
        pub_key = readfile(f'{Configure.KEY_DIR}/id_rsa.pub').strip()
        self.manager_public_key = pub_key
        Shell.execute('rm', path_expand(f'{Configure.KEY_DIR}/id_rsa'))
        Shell.execute('rm', path_expand(f'{Configure.KEY_DIR}/id_rsa.pub'))
        return priv_key, pub_key



import time

from cloudmesh.burn.burner.BurnerABC import AbstractBurner
from cloudmesh.burn.image import Image
from cloudmesh.burn.raspberryos.cmdline import Cmdline
from cloudmesh.burn.raspberryos.runfirst import Runfirst
from cloudmesh.burn.sdcard import SDCard
from cloudmesh.burn.usb import USB
from cloudmesh.common.console import Console
from cloudmesh.common.parameter import Parameter
from cloudmesh.common.util import yn_choice, readfile, banner
from cloudmesh.inventory.inventory import Inventory


class Burner(AbstractBurner):
    """
    Burner uses a cloudmesh inventory to create a RaspberryOS cluster

    Inventory should contain information on manager and workers
    """
    def __init__(self, inventory=None):
        # Get inventory
        if inventory is None:
            inv = Inventory()
        else:
            inv = Inventory(filename=inventory)

        # Find managers and workers
        managers = inv.find(service='manager')
        workers = inv.find(service='worker')
        self.inventory = inv

        # No inherenet need to distinguish the configs by service
        configs = managers + workers
        # Create dict for them for easy lookup
        self.configs = dict((config['host'], config) for config in configs)
        self.get_images()

    def get_images(self):
        """
        Downloads all tags found in self.configs
        """
        tags = set()
        for config in self.configs.values():
            try:
                tags.add(config['tag'])
            except KeyError as e:
                Console.warning(f'Could not find tag for {config["host"]}. Skipping')

        banner("Downloading Images", figlet=True)

        image = Image()

        for tag in tags:
            Console.info(f'Attempting to download {tag}')
            res = image.fetch(tag=[tag])
            if not res:
                Console.error('Failed Image Fetch.')
                raise Exception('Failed Image Fetch')

    def cluster(self, arguments=None):
        raise NotImplementedError

    def burn(self,
             name=None,
             device=None,
             verbose=False,
             password=None,
             ssid=None,
             wifipasswd=None,
             country=None):
        """
        Given the name of a config, burn device with RaspberryOS and configure properly
        """
        if device is None:
            Console.error('Device not specified')
            return
        if name is None:
            Console.error('Name to burn is not specified')
            return
        if name not in self.configs:
            Console.error(f'Could not find {name} in Inventory. Is the service column marked as "manager" or "worker"?')
            return

        config = self.configs[name]
        sdcard = SDCard(card_os="raspberry")

        try:
            USB.check_for_readers()
        except Exception as e:
            print()
            Console.error(e)
            print()
            return ""

        # Confirm card is inserted into device path
        if not yn_choice(f'Is the card to be burned for {name} inserted?'):
            if not yn_choice(f"Please insert the card to be burned for {name}. "
                             "Type 'y' when done or 'n' to terminante"):
                Console.error("Terminating: User Break")
                return ""

        Console.info(f'Burning {name}')
        sdcard.format_device(device=device, yes=True)
        sdcard.unmount(device=device)
        sdcard.burn_sdcard(tag=config['tag'], device=device, yes=True)
        sdcard.mount(device=device, card_os="raspberry")

        # Read and write cmdline.txt
        cmdline = Cmdline()
        # Reading will create the proper script in the cmdline instance
        # No extra work needed
        cmdline.read(filename=f'{sdcard.boot_volume}/cmdline.txt')
        cmdline.write(filename=f'{sdcard.boot_volume}/cmdline.txt')
        # print(cmdline.get())

        # Build the proper runfrist.sh
        runfirst = Runfirst()
        runfirst.set_hostname(config['host'])
        other_hosts, other_ips = self._get_hosts_for(name=config['host'])
        runfirst.set_hosts(names=other_hosts, ips=other_ips)
        if config['ip']:
            # config['router'] and config['dns'] are allowed to be empty String or None to skip its config
            # Default column in inventory is empty string
            runfirst.set_static_ip(ip=config['ip'], router=config['router'], dns=config['dns'])

        if password:
            runfirst.set_password(password=password)

        runfirst.set_locale(timezone=config['timezone'], locale=config['locale'])
        if ssid and 'wifi' in config['services']:
            runfirst.set_wifi(ssid, wifipasswd, country=country)

        runfirst.set_key(key=readfile(config['keyfile']).strip())
        if 'bridge' in config['services']:
            runfirst.enable_bridge()

        runfirst.get(verbose=verbose)
        runfirst.write(filename=f'{sdcard.boot_volume}/{Runfirst.SCRIPT_NAME}')

        time.sleep(1)  # Sleep for 1 seconds to give ample time for writing to finish
        sdcard.unmount(device=device, card_os="raspberry")
        Console.ok(f'Burned {name}')

        return

    def inventory(self, arguments=None):
        return self.inventory

    def multi_burn(self,
                   names=None,
                   devices=None,
                   verbose=False,
                   password=None,
                   ssid=None,
                   wifipasswd=None,
                   country=None):
        """
        Given multiple names, burn them
        """
        if devices is None:
            Console.error('Device not specified.')
            return
        if names is None:
            Console.error('Names to burn not specified')
            return

        names = Parameter.expand(names)
        devices = Parameter.expand(devices)

        if len(devices) > 1:
            Console.error('We do not yet support burning on multiple devices')
            return

        for name in names:
            self.burn(
                name=name,
                device=devices[0],
                verbose=verbose,
                password=password,
                ssid=ssid,
                wifipasswd=wifipasswd,
                country=None
            )
        Console.ok('Finished burning all cards')

    def _get_hosts_for(self, name=None):
        """
        Given a name, return the list of hostnames and ips to go into /etc/hosts
        """
        if name is None:
            raise Exception('name is None')
        if self.configs is None:
            raise Exception('no configs supplied yet')

        host_names = []
        ips = []

        for host_name in self.configs:
            if name != host_name and self.configs[host_name]['ip']:
                host_names += [host_name]
                ips += [self.configs[host_name]['ip']]
        return host_names, ips
